import asyncio
from dataclasses import dataclass
import sys
if sys.version_info < (3, 11):
    from typing_extensions import Any, NotRequired, TypedDict, List, TextIO, Tuple, cast
else:
    from typing import Any, NotRequired, TypedDict, List, TextIO, Tuple, cast
from datetime import datetime
from aioquic.quic.configuration import QuicConfiguration
from aioquic.quic.logger import QuicLogger
from aioquic.tls import CipherSuite, SessionTicket

@dataclass
class QuicBidirectionalStream:
    reader: asyncio.StreamReader
    writer: asyncio.StreamWriter
    lock: asyncio.Lock
    
CipherSuiteType = CipherSuite | str | int

def toCipherSuite(value: CipherSuiteType) -> CipherSuite:
    if isinstance(value, CipherSuite):
        return value
    if isinstance(value, int):
        return CipherSuite(value)
    return CipherSuite[value]

class SessionTicketDict(TypedDict):
    """
    A TLS session ticket for session resumption.
    """

    age_add: int
    cipher_suite: CipherSuiteType
    not_valid_after: str | float | datetime
    not_valid_before: str | float | datetime
    resumption_secret: bytes
    server_name: str
    ticket: bytes

    max_early_data_size: NotRequired[int | None]
    other_extensions: NotRequired[List[Tuple[int, bytes]]]
    
def toDatetime(value: str | float | datetime) -> datetime:
    if isinstance(value, datetime):
        return value
    if isinstance(value, float):
        return datetime.fromtimestamp(value)
    return datetime.fromisoformat(value)
    
def toSessionTicket(value: SessionTicket | SessionTicketDict) -> SessionTicket:
    if isinstance(value, SessionTicket):
        return value
    value = value.copy()
    value['cipher_suite'] = toCipherSuite(value['cipher_suite'])
    value['not_valid_after'] = toDatetime(value['not_valid_after'])
    value['not_valid_before'] = toDatetime(value['not_valid_before'])
    return SessionTicket(**cast(dict, value))
    
    
class QuicConfigurationDict(TypedDict):
    """
    A QUIC configuration.
    """

    alpn_protocols: NotRequired[List[str] | None]
    """
    A list of supported ALPN protocols.
    """

    connection_id_length: NotRequired[int]
    """
    The length in bytes of local connection IDs.
    """

    idle_timeout: NotRequired[float]
    """
    The idle timeout in seconds.

    The connection is terminated if nothing is received for the given duration.
    """

    is_client: NotRequired[bool]
    """
    Whether this is the client side of the QUIC connection.
    """

    max_data: NotRequired[int]
    """
    Connection-wide flow control limit.
    """

    max_stream_data: NotRequired[int]
    """
    Per-stream flow control limit.
    """

    quic_logger: NotRequired[QuicLogger | None]
    """
    The :class:`~aioquic.quic.logger.QuicLogger` instance to log events to.
    """

    secrets_log_file: NotRequired[TextIO | None]
    """
    A file-like object in which to log traffic secrets.

    This is useful to analyze traffic captures with Wireshark.
    """

    server_name: NotRequired[str | None]
    """
    The server name to send during the TLS handshake the Server Name Indication.

    .. note:: This is only used by clients.
    """

    session_ticket: NotRequired[SessionTicket | SessionTicketDict | None]
    """
    The TLS session ticket which should be used for session resumption.
    """

    cadata: NotRequired[bytes | None]
    cafile: NotRequired[str | None]
    capath: NotRequired[str | None]
    certificate: NotRequired[Any | None]
    certificate_chain: NotRequired[List[Any]]
    cipher_suites: NotRequired[List[CipherSuiteType] | None]
    initial_rtt: NotRequired[float]
    max_datagram_frame_size: NotRequired[int | None]
    private_key: NotRequired[Any | None]
    quantum_readiness_test: NotRequired[bool]
    supported_versions: NotRequired[List[int]]
    verify_mode: NotRequired[int | None]
    
def toQuicConfiguration(value: QuicConfiguration | QuicConfigurationDict) -> QuicConfiguration:
    if isinstance(value, QuicConfiguration):
        return value
    value = value.copy()
    session_ticket = value.get('session_ticket')
    if session_ticket is not None:
        value['session_ticket'] = toSessionTicket(session_ticket)
    cipher_suites = value.get('cipher_suites')
    if cipher_suites is not None:
        value['cipher_suites'] = [toCipherSuite(s) for s in cipher_suites]
    return QuicConfiguration(**cast(dict, value))


import logging
import uuid
from dataclasses import dataclass, field
from typing import Any, Callable, Literal, cast, overload

from aiortc import MediaStreamTrack

from .common import EventDispatcher, OwtEvent
from .constants import AudioCodec, AudioSourceInfo, VideoCodec, VideoSourceInfo
from .portal_types import MediaInfo, StreamInfo
from .publication import (AudioCodecParameters, AudioPublicationSettings,
                          PublicationSettings, Resolution,
                          VideoCodecParameters, VideoPublicationSettings)
from .sub_types import (AudioSubscriptionCapabilities,
                        SubscriptionCapabilities,
                        VideoSubscriptionCapabilities)
from .quic_types import QuicBidirectionalStream
from .log import configLogger

logger = configLogger(logger=logging.getLogger('owt-stream'))

@dataclass
class StreamSourceInfo:
    """
    Audio source info or video source info could be undefined if a stream does not have audio/video track.
    """
    
    audio: AudioSourceInfo | None = None
    """
    Audio source info. Accepted values are: "mic", "screen-cast", "file", "mixed" or undefined.
    """
    video: VideoSourceInfo | None = None
    """
    Video source info. Accepted values are: "camera", "screen-cast", "file", "mixed" or undefined.
    """
    data: bool = False
    """
    Indicates whether it is data. Accepted values are boolean.
    """

@dataclass
class Stream(EventDispatcher):
    tracks: list[MediaStreamTrack]
    quic_stream: QuicBidirectionalStream | None
    source: StreamSourceInfo
    attributes: dict = field(default_factory=dict)
    
    def __post_init__(self):
        super().__init__()
    
    def hasAudio(self) -> bool:
        for track in self.tracks:
            if track.kind == 'audio':
                return True
        return False
    
    def hasVideo(self) -> bool:
        for track in self.tracks:
            if track.kind == 'video':
                return True
        return False
    
    def hasData(self) -> bool:
        return self.quic_stream is not None
    
    @property
    def audio_num(self) -> int:
        i = 0
        for track in self.tracks:
            if track.kind == 'audio':
                i += 1
        return i
    
    @property
    def audioTracks(self) -> list[MediaStreamTrack]:
        return [track for track in self.tracks if track.kind == 'audio']
    
    @property
    def audioTrack(self) -> MediaStreamTrack | None:
        for track in self.tracks:
            if track.kind == 'audio':
                return track
        return None
    
    @property
    def video_num(self) -> int:
        i = 0
        for track in self.tracks:
            if track.kind == 'video':
                i += 1
        return i
    
    @property
    def videoTracks(self) -> list[MediaStreamTrack]:
        return [track for track in self.tracks if track.kind == 'video']
    
    @property
    def videoTrack(self) -> MediaStreamTrack | None:
        for track in self.tracks:
            if track.kind == 'video':
                return track
        return None
    
    @property
    def data_stream(self) -> QuicBidirectionalStream:
        assert self.quic_stream, f'Only data stream has quic stream.'
        return self.quic_stream

@dataclass
class LocalStream(Stream):
    id: str = field(default_factory=lambda: f'{uuid.uuid1()}')
    
@dataclass
class RemoteStream(Stream):
    id: str = field(default_factory=lambda: f'{uuid.uuid1()}')
    origin: str | None = None
    """
    ID of the remote endpoint who published this stream.
    """
    settings: PublicationSettings | None = None
    """
    Original settings for publishing this stream. This property is only valid in conference mode.
    """
    extraCapabilities: SubscriptionCapabilities | None = None
    """
    Extra capabilities remote endpoint provides for subscription. Extra
    capabilities don't include original settings. This property is only valid
    in conference mode.
    """
    
    @overload
    def addEventListener(self, eventType: Literal['ended'], listener: Callable[[OwtEvent], Any], once: bool = False, seq: bool = True) -> None: ...
    @overload
    def addEventListener(self, eventType: Literal['activeaudioinputchange'], listener: Callable[["ActiveAudioInputChangeEvent"], Any], once: bool = False, seq: bool = True) -> None: ...
    @overload
    def addEventListener(self, eventType: Literal['layoutchange'], listener: Callable[["LayoutChangeEvent"], Any], once: bool = False, seq: bool = True) -> None: ...
    def addEventListener(self, eventType: str, listener: Callable[[Any], Any], once: bool = False, seq: bool = True) -> None:
        return super().addEventListener(eventType, listener, once=once, seq=seq)
    
def convertToPublicationSettings(mediaInfo: MediaInfo | None):
    if mediaInfo is None:
        return PublicationSettings()
    audio: list[AudioPublicationSettings] = []
    video: list[VideoPublicationSettings] = []
    for track in mediaInfo['tracks']:
        if track['type'] == 'audio':
            format = track.get('format')
            if format:
                audio_codec = AudioCodecParameters(
                    name=cast(AudioCodec, format.get('codec')),
                    channelCount=format.get('channelNum'),
                    clockRate=format.get('sampleRate'),
                )
            else:
                audio_codec = None
            audio_publication_settings = AudioPublicationSettings(codec=audio_codec)
            audio_publication_settings._trackId = track.get('id')
            audio.append(audio_publication_settings)
        elif track['type'] == 'video':
            format = track.get('format')
            if format:
                video_codec = VideoCodecParameters(
                    name=cast(VideoCodec, format.get('codec')),
                    profile=format.get('profile'),
                )
            else:
                video_codec = None
            parameters = track.get('parameters')
            if parameters:
                r = parameters.get('resolution')
                resolution = Resolution(width=r['width'], height=r['height']) if r else None
                framerate = parameters.get('framerate')
                bitrate = parameters.get('bitrate')
                keyFrameInterval = parameters.get('keyFrameInterval')
            else:
                resolution = None
                framerate = None
                bitrate = None
                keyFrameInterval = None
            rid = track.get('rid')
                
            video_publication_settings = VideoPublicationSettings(
                codec=video_codec,
                resolution=resolution,
                frameRate=framerate,
                bitrate=bitrate,
                keyFrameInterval=keyFrameInterval,
                rid=rid,
            )
            video_publication_settings._trackId = track.get('id')
            video.append(video_publication_settings)
    return PublicationSettings(audio=audio, video=video)

def extractBitrateMultiplier(input: float | str) -> float:
    """
    Extract bitrate multiplier from a string like "x0.2".

    Args:
        input (float | str): bitrate multiplier
    Return:
        The float number after "x".
    """
    if isinstance(input, str):
        if not input.startswith('x'):
            logger.warning(f'Invalid bitrate multiplier input: {input}')
            return 0
        return float(input[1:])
    else:
        return input

def convertToSubscriptionCapabilities(mediaInfo: MediaInfo | None):
    if mediaInfo is None:
        return SubscriptionCapabilities()
    audio = None
    video = None
    for track in mediaInfo['tracks']:
        if track['type'] == 'audio':
            audio_codecs: list[AudioCodecParameters] = []
            optional = track.get('optional')
            if optional:
                format = optional.get('format')
                if format:
                    for audio_codec_info in format:
                        audio_codec = AudioCodecParameters(
                            name=cast(AudioCodec, audio_codec_info.get('codec')),
                            channelCount=audio_codec_info.get('channelNum'),
                            clockRate=audio_codec_info.get('sampleRate'),
                        )
                        audio_codecs.append(audio_codec)
            audio_codecs.sort(key=lambda p: p.name)
            audio = AudioSubscriptionCapabilities(codecs=audio_codecs)
        elif track['type'] == 'video':
            video_codecs: list[VideoCodecParameters] = []
            resolutions: list[Resolution] = []
            bitrates: list[float] = []
            frameRates: list[float] = []
            keyFrameIntervals: list[float] = []
            optional = track.get('optional')
            if optional:
                format = optional.get('format')
                if format:
                    for video_codec_info in format:
                        video_codec = VideoCodecParameters(
                            name=cast(VideoCodec, video_codec_info.get('codec')),
                            profile=video_codec_info.get('profile'),
                        )
                        video_codecs.append(video_codec)
                parameters = optional.get('parameters')
                if parameters:
                    resolutions = [Resolution(width=r['width'], height=r['height']) for r in parameters.get('resolution') or []]
                    resolutions.sort(key=lambda r: (r.width, r.height))
                    bitrates = [extractBitrateMultiplier(b) for b in parameters.get('bitrate') or []]
                    if not bitrates or bitrates[-1] != 1.0:
                        bitrates.append(1.0)
                    bitrates.sort()
                    frameRates = [r for r in parameters.get('framerate') or []]
                    frameRates.sort()
                    keyFrameIntervals = [i for i in parameters.get('keyFrameInterval') or []]
                    keyFrameIntervals.sort()
            video_codecs.sort(key=lambda p: p.name)
            video = VideoSubscriptionCapabilities(
                codecs=video_codecs,
                resolutions=resolutions,
                frameRates=frameRates,
                bitrateMultipliers=bitrates,
                keyFrameIntervals=keyFrameIntervals,
            )
    return SubscriptionCapabilities(audio=audio, video=video)
    
@dataclass
class RemoteMixedStream(RemoteStream):
    
    def __init__(self, info: StreamInfo):
        assert info['type'] == 'mixed'
        super().__init__(id=info.get('id'), tracks=[], quic_stream=None, source=StreamSourceInfo(audio='mixed', video='mixed'))
        self.settings = convertToPublicationSettings(info.get('media'))
        self.extraCapabilities = convertToSubscriptionCapabilities(info.get('media'))
     
class StreamEvent(OwtEvent):
    stream: RemoteStream
    
    def __init__(self, type: str, stream: RemoteStream) -> None:
        super().__init__(type)
        self.stream = stream
        
class ActiveAudioInputChangeEvent(OwtEvent):
    activeAudioInputStreamId: str
    
    def __init__(self, type: str, activeAudioInputStreamId: str) -> None:
        super().__init__(type)
        self.activeAudioInputStreamId = activeAudioInputStreamId
        
class LayoutChangeEvent(OwtEvent):
    layout: dict
    
    def __init__(self, type: str, layout) -> None:
        super().__init__(type)
        self.layout = layout


import logging
import re
from typing import Literal

from .publication import AudioEncodingParameters, VideoEncodingParameters
from .log import configLogger

logger = configLogger(logging.getLogger('owt-utils'))

def findLineInRange(sdpLines: list[str], startLine: int, endLine: int, prefix: str, substr: str | None = None) -> int:
    realEndLine = endLine if endLine != -1 else len(sdpLines)
    for i in range(startLine, realEndLine):
        if sdpLines[i].startswith(prefix):
            if not substr or sdpLines[i].lower().find(substr.lower()) != -1:
                return i
    return -1

def findLine(sdpLines: list[str], prefix: str, substr: str | None = None) -> int:
    return findLineInRange(sdpLines, 0, -1, prefix, substr)

def findMLineRangeWithMID(sdpLines: list[str], mid: str) -> tuple[int, int] | tuple[None, None]:
    midLine = f'a=mid:{mid}'
    midIndex = findLine(sdpLines, midLine)
    if midIndex < 0:
        logging.debug('Specified MID is not found')
        return None, None
    while midIndex >= 0 and sdpLines[midIndex] != midLine:
        midIndex = findLineInRange(sdpLines, midIndex, -1, midLine)
    if midIndex >= 0:
        nextMLineIndex = findLineInRange(sdpLines, midIndex, -1, 'm=')
        mLineIndex = -1
        for i in reversed(range(midIndex)):
            if sdpLines[i].find('m=') >= 0:
                mLineIndex = i
                break
        if mLineIndex >= 0:
            return mLineIndex, nextMLineIndex
    return None, None

def getCodecPayloadTypeFromLine(sdpLine: str) -> str | None:
    m = re.search(r'a=rtpmap:(\d+) [a-zA-Z0-9-]+\/\d+', sdpLine)
    if m is not None:
        return m.group(1)
    else:
        return None

def getCodecPayloadType(sdpLines: list[str], codec: str) -> str | None:
    index = findLine(sdpLines, 'a=rtpmap', codec)
    if index != -1:
        return getCodecPayloadTypeFromLine(sdpLines[index])
    else:
        return None
    
def parseFmtpLine(fmtpLine: str) -> tuple[str, dict[str, str]] | tuple[None, None]:
    """
    Split an fmtp line into an tuple including 'pt' and 'params'.

    Args:
        fmtpLine (str): an fmtp line

    Returns:
        tuple[str, list[str]]: 'pt' and 'params'
    """
    space_pos = fmtpLine.find(' ')
    assert space_pos != -1
    key_values = fmtpLine[space_pos + 1:].split(';')
    result = re.search(r'a=fmtp:(\d+)', fmtpLine)
    if not result:
        return None, None
    pt = result.group(1)
    params = { key_value[0]: key_value[1] for key_value in (key_value.split('=') for key_value in key_values) if len(key_value) == 2 }
    return pt, params
    
        
def writeFmtpLine(pt: str, params: dict[str, str]) -> str | None:
    key_values = [f'{key}={value}' for key, value in params.items()]
    if len(key_values) == 0:
        return None
    return f'a=fmtp:{pt} {";".join(key_values)}'

def findFmtpLine(sdpLines: list[str], codec: str) -> int:
    payload = getCodecPayloadType(sdpLines, codec)
    return findLine(sdpLines, 'a=fmtp:' + f'{payload}') if payload is not None else -1
        

def setCodecParam(sdp: str, codec: str, param, value, mid: str | None = None) -> str:
    sdpLines = sdp.split('\r\n')
    headLines = None
    tailLines = None
    if mid:
        start, end = findMLineRangeWithMID(sdpLines, mid)
        if start is not None:
            headLines = sdpLines[0:start]
            tailLines = sdpLines[end:]
            sdpLines = sdpLines[start:end]
    if len(sdpLines) <= 1:
        sdpLines = sdp.split('\n')
    fmtpLineIndex = findFmtpLine(sdpLines, codec)
    if fmtpLineIndex == -1:
        index = findLine(sdpLines, 'a=rtpmap', codec)
        if index == -1:
            return sdp
        payload = getCodecPayloadTypeFromLine(sdpLines[index])
    else:
        pt, params = parseFmtpLine(sdpLines[fmtpLineIndex])
        assert pt is not None
        assert params is not None
        params[param] = f'{value}'
        fmtpline = writeFmtpLine(pt, params)
        assert fmtpline is not None
        sdpLines[fmtpLineIndex] = fmtpline
    if headLines:
        assert tailLines
        sdpLines = [*headLines, *sdpLines, *tailLines]
    sdp = '\r\n'.join(sdpLines)
    return sdp
# Following codecs will not be removed from SDP event they are not in the
# user-specified codec list.
audioCodecAllowList = ['CN', 'telephone-event']
videoCodecAllowList = ['red', 'ulpfec', 'flexfec']

def removeCodecFramALine(sdpLines: list[str], payload: str) -> list[str]:
    new_sdp_lines: list[str] = []
    for line in sdpLines:
        if not re.search(f'a=(rtpmap|rtcp-fb|fmtp):{re.escape(payload)}\\s', line):
            new_sdp_lines.append(line)
    return new_sdp_lines

def setCodecOrder(mLine: str, payloads: list[str]) -> str:
    """
    Returns a new m= line with the specified codec order.

    Args:
        mLine (str): m= line
        payloads (list[str]): payloads

    Returns:
        str: new m= line
    """
    elements = mLine.split(' ')
    # Just copy the first three parameters; codec order starts on fourth.
    new_line = elements[0:3]
    new_line = [*new_line, *payloads]
    return ' '.join(new_line)

def appendRtxPayloads(sdpLines: list[str], payloads: list[str]) -> list[str]:
    new_payloads = [*payloads]
    for payload in payloads:
        index = findLine(sdpLines, 'a=fmtp', f'apt={payload}')
        if index != -1:
            pt, _ = parseFmtpLine(sdpLines[index])
            assert pt is not None
            new_payloads.append(pt)
    return new_payloads

def reorderCodecs(sdp: str, type: Literal['audio', 'video'], codecs: list[str], mid: str | None = None) -> str:
    if not codecs:
        return sdp
    codecs = [*codecs, *(audioCodecAllowList if type == 'audio' else videoCodecAllowList)]
    sdpLines = sdp.split('\r\n')
    headLines = None
    tailLines = None
    if mid:
        start, end = findMLineRangeWithMID(sdpLines, mid)
        if start is not None:
            headLines = sdpLines[0:start]
            tailLines = sdpLines[end:]
            sdpLines = sdpLines[start:end]
    mLineIndex = findLine(sdpLines, 'm=', type)
    if mLineIndex == -1:
        return sdp
    originPayloads = sdpLines[mLineIndex].split(' ')
    originPayloads = originPayloads[3:]
    payloads: list[str] = []
    for codec in codecs:
        i = 0
        while i < len(sdpLines):
            index = findLineInRange(sdpLines, i, -1, 'a=rtpmap', codec)
            if index != -1:
                payload = getCodecPayloadTypeFromLine(sdpLines[index])
                if payload:
                    payloads.append(payload)
                    i = index
            i += 1
    payloads = appendRtxPayloads(sdpLines, payloads)
    sdpLines[mLineIndex] = setCodecOrder(sdpLines[mLineIndex], payloads)
    for payload in originPayloads:
        if payloads.count(payload) == 0:
            sdpLines = removeCodecFramALine(sdpLines, payload)
    if headLines:
        assert tailLines is not None
        sdpLines = [*headLines, *sdpLines, *tailLines]
    sdp = '\r\n'.join(sdpLines)
    return sdp

def setMaxBitrate(sdp: str, encodingParametersList: list[AudioEncodingParameters] | list[VideoEncodingParameters], mid: str | None):
    for p in encodingParametersList:
        if p.maxBitrate is not None and p.codec is not None:
            sdp = setCodecParam(sdp, p.codec.name, 'x-google-max-bitrate', p.maxBitrate, mid)
    return sdp


from dataclasses import dataclass

from .constants import AudioCodec, VideoCodec, VideoProfile


@dataclass
class AudioCodecParameters:
    """
    Codec parameters for an audio track.
    """
    name: AudioCodec
    """
    Name of a codec. Please a value in AudioCodec. However,
    some functions do not support all the values in AudioCodec.
    """
    channelCount: int | None = None
    """
    Numbers of channels for an audio track.
    """
    clockRate: float | None = None
    """
    The codec clock rate expressed in Hertz.
    """
    

@dataclass
class AudioEncodingParameters:
    """
    Encoding parameters for sending an audio track.
    """
    codec: AudioCodecParameters | None = None
    maxBitrate: float | None = None

@dataclass
class VideoCodecParameters:
    """
    Codec parameters for a video track.
    """
    name: VideoCodec
    """
    Name of a codec.Please a value in VideoCodec.However,
    some functions do not support all the values in VideoCodec.
    """
    profile: VideoProfile | None = None
    """
    The profile of a codec. Profile may not apply to all codecs.
    If codec not equals "h264", should be None
    """

@dataclass
class VideoEncodingParameters:
    """
    Encoding parameters for sending a video track.
    """
    
    codec: VideoCodecParameters | None = None
    maxBitrate: float | None = None
    """
    Max bitrate expressed in kbps.
    """


import asyncio
import collections.abc as abc
import inspect
import json
import time
from collections.abc import Iterable
from threading import Lock
from contextlib import AsyncExitStack
from typing import (Any, Awaitable, Callable, Generic, Literal, TypeAlias,
                    TypeVar, cast, get_origin, get_type_hints)

import aiohttp
import six

I = TypeVar('I')
O = TypeVar('O')

def remove_none_value(input: I, recursion: bool = False, check_circular: bool = True, ref_set: set[int] | None = None) -> I:
    if check_circular:
        ref_set = ref_set if ref_set is not None else set()
        if id(input) in ref_set:
            return input
        ref_set.add(id(input))
    if isinstance(input, dict):
        keys = list(input.keys())
        for k in keys:
            v = input[k]
            if v is None:
                del input[k]
            elif recursion:
                remove_none_value(v, recursion=recursion, check_circular=check_circular, ref_set=ref_set)
    elif isinstance(input, Iterable) and not isinstance(input, six.string_types):
        for e in input:
            remove_none_value(e, recursion=recursion, check_circular=check_circular, ref_set=ref_set)
                
    return input

def chain_future(
    future: asyncio.Future[I],
    then: Callable[[I], O] | Callable[[I], Awaitable[O]] | None = None,
    catch: Callable[[BaseException], Any] | Callable[[BaseException], Awaitable] | None = None
) -> asyncio.Future[O]:
    result = asyncio.Future(loop=future.get_loop())
    def callback(f: asyncio.Future[I]):
        if result.done():
            return
        if f.cancelled():
            result.cancel()
            return
        e = f.exception()
        if e is None:
            if then is not None:
                out = then(f.result())
                if inspect.isawaitable(out):
                    f_out = asyncio.ensure_future(out)
                    link_future(result, f_out)
                else:
                    result.set_result(out)
            else:
                result.set_result(f.result())
        else:
            if catch is not None:
                out = catch(e)
                if inspect.isawaitable(out):
                    f_out = asyncio.ensure_future(out)
                    link_future(result, f_out)
                else:
                    result.set_result(out)
            else:
                result.set_exception(e)
    future.add_done_callback(callback)
    return result

def link_future(this_future: asyncio.Future[I], target_future: asyncio.Future[I]) -> None:
    def on_done(f: asyncio.Future):
        if f.cancelled():
            this_future.cancel()
        elif not this_future.done():
            e = f.exception()
            if e:
                this_future.set_exception(e)
            else:
                this_future.set_result(f.result())
    target_future.add_done_callback(on_done)

class AsyncExceptionCollector:
    loop: asyncio.AbstractEventLoop | None
    lock: Lock
    future_lock: Lock
    listeners_waiter: asyncio.Future[Literal[True]] | None
    exceptions_waiter: asyncio.Future[Literal[True]] | None
    listeners_counter: int
    exceptions_counter: int
    future: asyncio.Future[None]
    pending_exceptions: list[BaseException]
    listeners: list["AsyncExceptionCollectorListenerWrapper"]
    
    Callback = Callable[[BaseException], bool] | Callable[[BaseException], Awaitable[bool]]
    
    def __init__(self, loop: asyncio.AbstractEventLoop | None = None) -> None:
        self.loop = loop
        self.future = asyncio.Future(loop=loop)
        self.listeners = []
        self.pending_exceptions = []
        self.lock = Lock()
        self.future_lock = Lock()
        self.listeners_waiter = None
        self.exceptions_waiter = None
        self.listeners_counter = 0
        self.exceptions_counter = 0
    
    def collect(self, consume: bool = True):
        return AsyncExceptionCollectorContextManager(self, consume)
    
    async def lock_listeners(self, timeout: float = 0):
        with self.lock:
            assert self.listeners_counter >= 0
            if self.listeners_counter > 0 or self.exceptions_counter == 0:
                self.listeners_counter += 1
                return True
            elif not self.listeners_waiter:
                waiter = asyncio.Future(loop=self.loop)
                self.listeners_waiter = waiter
            else:
                waiter = self.listeners_waiter
        t0 = time.time()
        if timeout > 0:
            try:
                await asyncio.wait_for(waiter, timeout=timeout)
            except asyncio.TimeoutError:
                return False
        else:
            await waiter
        t1 = time.time()
        timeout = timeout - (t1 - t0)
        return await self.lock_listeners(timeout=timeout)
        
        
    def unlock_listeners(self):
        full_unlock = False
        pending_future: asyncio.Future[Literal[True]] | None = None
        with self.lock:
            if self.listeners_counter <= 0:
                raise ValueError('Must unlock when has locked.')
            assert self.listeners_waiter is None
            self.listeners_counter -= 1
            if self.listeners_counter == 0:
                full_unlock = True
                pending_future = self.exceptions_waiter
        if full_unlock and pending_future:
            pending_future.set_result(True)
    
        
    async def lock_exceptions(self, timeout: float = 0):
        with self.lock:
            assert self.exceptions_counter >= 0
            if self.exceptions_counter > 0 or self.listeners_counter == 0:
                self.exceptions_counter += 1
                return True
            elif not self.exceptions_waiter:
                waiter = asyncio.Future(loop=self.loop)
                self.exceptions_waiter = waiter
            else:
                waiter = self.exceptions_waiter
        t0 = time.time()
        if timeout > 0:
            try:
                await asyncio.wait_for(waiter, timeout=timeout)
            except asyncio.TimeoutError:
                return False
        else:
            await waiter
        t1 = time.time()
        timeout = timeout - (t1 - t0)
        return await self.lock_listeners(timeout=timeout)
    
    def unlock_exceptions(self):
        full_unlock = False
        pending_future: asyncio.Future[Literal[True]] | None = None
        with self.lock:
            if self.exceptions_counter <= 0:
                raise ValueError('Must unlock when has locked.')
            assert self.exceptions_waiter is None
            self.exceptions_counter -= 1
            if self.exceptions_counter == 0:
                full_unlock = True
                pending_future = self.listeners_waiter
        if full_unlock and pending_future:
            pending_future.set_result(True)
            
    async def invoke_listenr(self, listener: "AsyncExceptionCollectorListenerWrapper", e: BaseException):
        try:
            res = listener.invoke(e, self)
        except BaseException as le:
            self.pending_exceptions.append(le)
            res = False
        if inspect.isawaitable(res):
            res = await res
        if res:
            return True
        else:
            return False
    
    async def add_listener(self, listener: "AsyncExceptionCollectorListener") -> "AsyncExceptionCollector":
        await self.lock_listeners()
        l = next((l for l in self.listeners if l.listener == listener), None)
        if l:
            return self
        real_listener = AsyncExceptionCollectorListenerWrapper(listener)
        try:
            if self.future.done() and not self.future.cancelled():
                e = self.future.exception()
                if e:
                    consumed = await self.invoke_listenr(real_listener, e)
                    if consumed:
                        raise RuntimeError(f'The collector has stopped, it\'s impossible to consume the exception: {e}')
            self.listeners.append(real_listener)
            return self
        finally:
            self.unlock_listeners()
    
    def remove_listener(self, listener: "AsyncExceptionCollectorListener") -> bool:
        l = next((l for l in self.listeners if l.listener == listener), None)
        if l:
            try:
                self.listeners.remove(l)
                return True
            except ValueError:
                return False
        else:
            return False
        
    async def add_exception(self, e: BaseException) -> "AsyncExceptionCollector":
        await self.lock_exceptions()
        try:
            consume = False
            for listener in self.listeners:
                try:
                    res = listener.invoke(e, self)
                except BaseException as le:
                    self.pending_exceptions.append(le)
                    res = False
                if inspect.isawaitable(res):
                    try:
                        res = await res
                    except BaseException as le:
                        self.pending_exceptions.append(le)
                        res = False
                if res:
                    consume = True
                    break
            if not consume:
                with self.future_lock:
                    if self.future.done():
                        self.pending_exceptions.append(e)
                    else:
                        self.future.set_exception(e)
            return self
        finally:
            self.unlock_exceptions()
            
    def cancel(self):
        return self.future.cancel()
    
    def error(self, e: BaseException):
        with self.future_lock:
            if not self.future.done():
                self.future.set_exception(e)
                
    def complete(self):
        with self.future_lock:
            if not self.future.done():
                self.future.set_result(None)
    
    async def wait_until_error(self, timeout: int = 0, loop: asyncio.AbstractEventLoop | None = None):
        future: asyncio.Future[None] = asyncio.Future(loop=loop or self.loop)
        lock = Lock()
        def on_done(f: asyncio.Future[None]):
            if f.cancelled():
                future.cancel()
            else:
                e = f.exception()
                with lock:
                    if not future.done():
                        if e:
                            future.set_exception(e)
                        else:
                            future.set_result(None)
        with self.future_lock:
            if not self.future.done():
                self.future.add_done_callback(on_done)
        if self.future.done():
            on_done(self.future)
                
        if timeout > 0:
            await asyncio.wait_for(future, timeout=timeout)
        else:
            await future
                    

class AsyncExceptionCollectorContextManager:
    collector: AsyncExceptionCollector
    consume: bool
    
    def __init__(self, collector: AsyncExceptionCollector, consume: bool = True) -> None:
        self.collector = collector
        self.consume = consume
        
    def __enter__(self):
        raise NotImplementedError('AsyncExceptionCollectorContextManager is an async context manager, please use it by "async with" instead of "with".')
    
    def __exit__(self):
        return False
    
    async def __aenter__(self):
        return self.collector
    
    async def __aexit__(self, exc_type, exc_value, exc_tb):
        if exc_value:
            await self.collector.add_exception(exc_value)
        return self.consume
    
AsyncExceptionCollectorListener: TypeAlias = Callable[[BaseException, AsyncExceptionCollector], Any] | Callable[[BaseException], Any] | Callable[[], Any]

class AsyncExceptionCollectorListenerWrapper:
    listener: AsyncExceptionCollectorListener
    type_hints: dict[str, Any]
    exception_type: type[BaseException] | None
    accept_collector: bool
    exception_var_name: str | None
    collector_var_name: str | None
    reverse_order: bool
    valid: bool
    
    def __init__(self, listener: AsyncExceptionCollectorListener) -> None:
        self.listener = listener
        self.type_hints = self.get_type_hints()
        self.exception_type = None
        self.accept_collector = False
        self.exception_var_name = None
        self.collector_var_name = None
        self.reverse_order = False
        self.valid = False
        sig = inspect.signature(self.listener)
        self.analyze_signature(sig)
        if not self.valid:
            raise ValueError(f'Invalid input listener: {sig}')
                
    def is_collector_name(self, name: str):
        return name in ['collector']
    
    def is_collector_type(self, p: inspect.Parameter):
        t = self.type_hints.get(p.name)
        e_list_type = self.get_type(t, AsyncExceptionCollector)
        return p.default == inspect.Parameter.empty and e_list_type is not None
    
    def is_exception_name(self, name: str):
        return name in ['e', 'exception']
    
    def calc_exception_type(self, p: inspect.Parameter):
        if p.default != inspect.Parameter.empty:
            return None
        t = self.type_hints.get(p.name)
        e_type = self.get_type(t, BaseException)
        return e_type
    
    def invoke(self, exception: BaseException, collector: AsyncExceptionCollector):
        if not self.exception_type or isinstance(exception, self.exception_type):
            args = []
            if self.exception_type and not self.exception_var_name:
                args.append(exception)
            if self.accept_collector and not self.collector_var_name:
                args.append(collector)
            if self.reverse_order:
                args.reverse()
            kw_args = {}
            if self.exception_type and self.exception_var_name:
                kw_args[self.exception_var_name] = exception
            if self.accept_collector and self.collector_var_name:
                kw_args[self.collector_var_name] = collector
            return self.listener(*args, **kw_args)
        else:
            return None
            
                
    def analyze_signature(self, sig: inspect.Signature):
        e_backable = False
        e_list_backable = False
        for p in sig.parameters.values():
            if p.kind == inspect.Parameter.POSITIONAL_ONLY or p.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD:
                if self.is_exception_name(p.name) and self.calc_exception_type(p):
                    if self.exception_type:
                        if e_backable:
                            self.exception_type = self.calc_exception_type(p)
                            self.exception_var_name = p.name if p.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD else None
                            self.reverse_order = self.accept_collector
                            e_backable = False
                        elif p.default == inspect.Parameter.empty:
                            self.exception_type = None
                            self.accept_collector = False
                            return
                    else:
                        self.exception_type = self.calc_exception_type(p)
                        self.exception_var_name = p.name if p.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD else None
                        self.reverse_order = self.accept_collector
                        e_backable = False
                elif self.is_collector_name(p.name) and self.is_collector_type(p):
                    if self.accept_collector:
                        if e_list_backable:
                            self.accept_collector = True
                            self.collector_var_name = p.name if p.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD else None
                            self.reverse_order = not self.exception_type
                            e_list_backable = False
                        elif p.default == inspect.Parameter.empty:
                            self.exception_type = None
                            self.accept_collector = False
                            return
                    else:
                        self.accept_collector = True
                        self.collector_var_name = p.name if p.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD else None
                        self.reverse_order = not self.exception_type
                        e_list_backable = False
                elif not self.exception_type and self.calc_exception_type(p):
                    self.exception_type = self.calc_exception_type(p)
                    self.exception_var_name = p.name if p.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD else None
                    self.reverse_order = self.accept_collector
                    e_backable = p.default != inspect.Parameter.empty
                elif not self.accept_collector and self.is_collector_type(p):
                    self.accept_collector = True
                    self.collector_var_name = p.name if p.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD else None
                    self.reverse_order = not self.exception_type
                    e_list_backable = p.default != inspect.Parameter.empty
                elif p.default == inspect.Parameter.empty:
                    self.exception_type = None
                    self.accept_collector = False
                    return
            elif p.kind == inspect.Parameter.VAR_POSITIONAL:
                if not self.exception_type:
                    self.exception_type = self.calc_exception_type(p)
                    self.exception_var_name = None
                    self.reverse_order = self.accept_collector
                    e_backable = True
                if not self.accept_collector:
                    self.accept_collector = True
                    self.collector_var_name = None
                    self.reverse_order = not self.exception_type
                    e_list_backable = True
            elif p.kind == inspect.Parameter.KEYWORD_ONLY:
                if self.is_exception_name(p.name) and self.calc_exception_type(p):
                    if self.exception_type:
                        if e_backable:
                            self.exception_type = self.calc_exception_type(p)
                            self.exception_var_name = p.name
                            e_backable = False
                        elif p.default == inspect.Parameter.empty:
                            self.exception_type = None
                            self.accept_collector = False
                            return
                    else:
                        self.exception_type = self.calc_exception_type(p)
                        self.exception_var_name = p.name
                        e_backable = False
                elif self.is_collector_name(p.name) and self.is_collector_type(p):
                    if self.accept_collector:
                        if e_list_backable:
                            self.accept_collector = True
                            self.collector_var_name = p.name
                            e_list_backable = False
                        elif p.default == inspect.Parameter.empty:
                            self.exception_type = None
                            self.accept_collector = False
                            return
                    else:
                        self.accept_collector = True
                        self.collector_var_name = p.name
                        e_list_backable = False
                elif p.default == inspect.Parameter.empty:
                    self.exception_type = None
                    self.accept_collector = False
                    return
            elif p.kind == inspect.Parameter.VAR_KEYWORD:
                if not self.exception_type:
                    self.exception_type = self.calc_exception_type(p)
                    self.exception_var_name = 'exception'
                    e_backable = True
                if not self.accept_collector:
                    self.accept_collector = True
                    self.collector_var_name = 'collector'
                    e_list_backable = True
        self.valid = True

                    
    def extract_fun(self):
        if not inspect.isroutine(self.listener) and isinstance(self.listener, abc.Callable) and hasattr(self.listener, '__call__'):
            return self.listener.__call__
        else:
            return self.listener
                    
    def get_type_hints(self) -> dict[str, Any]:
        fun = self.extract_fun()
        try:
            return get_type_hints(fun)
        except:
            return {}
                        
    def get_type(self, type: Any, target_type: Any):
        if type is None:
            return target_type
        if isinstance(type, TypeVar):
            if hasattr(type, '__bound__'):
                bound = getattr(type, '__bound__')
                return self.get_type(bound, target_type)
            else:
                return target_type
        else:
            o_type = get_origin(type)
            type = o_type if o_type is not None else type
            if inspect.isclass(type):
                return type if issubclass(type, target_type) else None
            else:
                return target_type
            
class NullAsyncContext:
    
    async def __aenter__(self):
        return self
    
    async def __aexit__(self, exc_type, exc_value, exc_tb):
        return False
    
R = TypeVar('R')
    
def ensure_future(callback: Callable[..., R] | Callable[..., Awaitable[R]], *args, loop: asyncio.AbstractEventLoop | None = None, **kwargs) -> asyncio.Future[R]:
    exc = None 
    try:
        res = callback(*args, **kwargs)
    except BaseException as e:
        res = None
        exc = e
    if exc is None and inspect.isawaitable(res):
        future = asyncio.ensure_future(res, loop=loop)
    else:
        future = asyncio.Future(loop=loop)
        if exc is None:
            future.set_result(res)
        else:
            future.set_exception(exc)
    return future

def real_attribute_name(inst: Any, attr_name: str):
    if attr_name.startswith('__'):
        return f'_{inst.__class__.__name__}{attr_name}'
    else:
        return attr_name

class Postponer:
    LOCK: Lock = Lock()
    lock: Lock
    
    future: asyncio.Future[None]
    
    def __init__(self) -> None:
        self.lock = Lock()
        self.future = asyncio.Future()
        self.future.set_result(None)
    
    def _chain_task(self, callback: Callable[[], Awaitable[None]]):
        async def then(_):
            await callback()
        with self.lock:
            self.future = chain_future(self.future, then=then)
            
    async def postpone(self, inst: Any, *methods, always: bool = True):
        stack = AsyncExitStack()
        try:
            for method in methods:
                if isinstance(method, six.string_types):
                    real_method_name = real_attribute_name(inst, method)
                    if not hasattr(inst, real_method_name) or not inspect.ismethod(getattr(inst, real_method_name)):
                        raise ValueError(f'{method} is not {inst.__class__.__name__}\'s instance method.')
                elif inspect.ismethod(method):
                    if not hasattr(method, '__org_name__'):
                        raise ValueError(f'{method} is not decorated by Postponer.enchance.')
                    real_method_name = real_attribute_name(inst, getattr(method, '__org_name__'))
                    if not hasattr(inst, real_method_name) or getattr(inst, real_method_name) != method:
                        raise ValueError(f'{method} is not {inst.__class__.__name__}\'s instance method.')
                else:
                    raise ValueError(f'{method} is not {inst.__class__.__name__}\'s instance method.')
                await stack.enter_async_context(Postponer.ContextManager(inst, self, real_method_name, always=always))
            return stack
        except:
            await stack.aclose()
            raise
    
            
    class ContextManager:
        inst: Any
        postponer: "Postponer"
        method: str
        always: bool
        
        def __init__(self, inst: Any, postponer: "Postponer", method: str, always: bool = True) -> None:
            self.inst = inst
            self.postponer = postponer
            self.method = method
            self.always = always
        
        async def __aenter__(self):
            Postponer.acquire_postponer(self.inst, self.method, self.postponer)
            return self.postponer
        
        async def __aexit__(self, exc_type, exc_value, exc_tb):
            Postponer.release_postponer(self.inst, self.method, self.postponer)
            if self.always or not exc_value:
                await self.postponer.future
            return False
                
    @staticmethod
    def get_postponers(inst, method_name: str, create: bool = False) -> list["Postponer"] | None:
        if hasattr(inst, '__postponers__'):
            postponers_map = cast(dict, getattr(inst, '__postponers__'))
        elif create:
            with Postponer.LOCK:
                if hasattr(inst, '__postponers__'):
                    postponers_map = getattr(inst, '__postponers__')
                else:
                    postponers_map = {}
                    setattr(inst, '__postponers__', postponers_map)
        else:
            return None
        postponers = postponers_map.get(method_name)
        if postponers:
            return postponers
        if create:
            with Postponer.LOCK:
                postponers = postponers_map.get(method_name)
                if not postponers:
                    postponers = postponers_map[method_name] = []
                return postponers
        else:
            return None
    
    @staticmethod
    def acquire_postponer(inst, method_name: str, postponer: "Postponer"):
        postponers = Postponer.get_postponers(inst, method_name, create=True)
        assert postponers is not None
        postponers.append(postponer)
        
    @staticmethod
    def release_postponer(inst, method_name: str, postponer: "Postponer"):
        postponers = Postponer.get_postponers(inst, method_name, create=False)
        assert postponers is not None
        postponers.remove(postponer)
    
    @staticmethod
    def enchance(method):
        if not inspect.iscoroutinefunction(method):
            raise ValueError(f'The Postponer.enchance must be decorated on the async method, but was decorated on method: {method.__qualname__}')
        if method.__qualname__.find('.') == -1:
            raise ValueError(f'The Postponer.enchance must be decorated on the instance method, but was decorated on method: {method.__qualname__}')
        async def wrapper(*args):
            self = args[0]
            postponers = Postponer.get_postponers(self, method.__name__,)
            if postponers:
                async def callback():
                    await method(*args)
                for postponer in postponers:
                    postponer._chain_task(callback=callback)
                return
            else:
                return await method(*args)
        wrapper.__org_name__ = method.__name__
        wrapper.__org_qualname__ = method.__qualname__
        return wrapper
            
        
                
async def create_token_from_owt_demo(url: str, username: str, role: str = 'presenter', room: str | None = None) -> str:
    async with aiohttp.ClientSession(connector=aiohttp.TCPConnector(verify_ssl=False)) as session:
        data = { 'room': room or '', 'user': username, 'role': role}
        async with session.post(f'{url}/tokens', headers={ 'Content-Type': 'application/json' }, data=json.dumps(data, separators=(',', ':'), ensure_ascii=True)) as r:
            return await r.text(encoding='utf-8')
        
        
if __name__ == '__main__':
    def t_f1(a):
        pass
    
    l = AsyncExceptionCollectorListenerWrapper(cast(AsyncExceptionCollectorListener, t_f1))
    assert l.exception_type == BaseException
    assert l.exception_var_name == 'a'
    assert not l.accept_collector
    assert l.collector_var_name is None
    assert not l.reverse_order
    assert l.valid
    
    def t_f2(a, b):
        pass
    
    l = AsyncExceptionCollectorListenerWrapper(cast(AsyncExceptionCollectorListener, t_f2))
    assert l.exception_type == BaseException
    assert l.exception_var_name == 'a'
    assert l.accept_collector
    assert l.collector_var_name == 'b'
    assert not l.reverse_order
    assert l.valid
    
    def t_f3(a: Exception, b: AsyncExceptionCollector):
        pass
    
    l = AsyncExceptionCollectorListenerWrapper(cast(AsyncExceptionCollectorListener, t_f3))
    assert l.exception_type == Exception
    assert l.exception_var_name == 'a'
    assert l.accept_collector
    assert l.collector_var_name == 'b'
    assert not l.reverse_order
    assert l.valid
    
    def t_f4(a: AsyncExceptionCollector, b: Exception):
        pass

    l = AsyncExceptionCollectorListenerWrapper(cast(AsyncExceptionCollectorListener, t_f4))
    assert l.exception_type == Exception
    assert l.exception_var_name == 'b'
    assert l.accept_collector
    assert l.collector_var_name == 'a'
    assert l.reverse_order
    assert l.valid
    
    def t_f5(a, b=1):
        pass
    
    l = AsyncExceptionCollectorListenerWrapper(cast(AsyncExceptionCollectorListener, t_f5))
    assert l.exception_type == BaseException
    assert l.exception_var_name == 'a'
    assert not l.accept_collector
    assert l.collector_var_name is None
    assert not l.reverse_order
    assert l.valid
    
    def t_f6(collector):
        pass
    
    l = AsyncExceptionCollectorListenerWrapper(cast(AsyncExceptionCollectorListener, t_f6))
    assert l.exception_type is None
    assert l.exception_var_name is None
    assert l.accept_collector
    assert l.collector_var_name == 'collector'
    assert l.valid
    
    def t_f7(*, collector, exception):
        pass
    
    l = AsyncExceptionCollectorListenerWrapper(cast(AsyncExceptionCollectorListener, t_f7))
    assert l.exception_type == BaseException
    assert l.exception_var_name == 'exception'
    assert l.accept_collector
    assert l.collector_var_name == 'collector'
    assert l.valid
    
    def t_f8(*args, **kwargs):
        pass
    
    l = AsyncExceptionCollectorListenerWrapper(cast(AsyncExceptionCollectorListener, t_f8))
    assert l.exception_type == BaseException
    assert l.exception_var_name is None
    assert l.accept_collector
    assert l.collector_var_name is None
    assert not l.reverse_order
    assert l.valid
    
    def t_f9(e: ValueError):
        pass
    
    l = AsyncExceptionCollectorListenerWrapper(cast(AsyncExceptionCollectorListener, t_f9))
    assert l.exception_type == ValueError
    assert l.exception_var_name == 'e'
    assert not l.accept_collector
    assert l.collector_var_name is None
    assert not l.reverse_order
    assert l.valid
    
    def t_f10(e: "ValueError"):
        pass
    
    l = AsyncExceptionCollectorListenerWrapper(cast(AsyncExceptionCollectorListener, t_f10))
    assert l.exception_type == ValueError
    assert l.exception_var_name == 'e'
    assert not l.accept_collector
    assert l.collector_var_name is None
    assert not l.reverse_order
    assert l.valid

    class t_f11:
        def __call__(self, es: AsyncExceptionCollector) -> Any:
            pass
    
    l = AsyncExceptionCollectorListenerWrapper(cast(AsyncExceptionCollectorListener, t_f11()))
    assert l.exception_type is None
    assert l.exception_var_name is None
    assert l.accept_collector
    assert l.collector_var_name == 'es'
    assert l.valid
    
    class t_f12:
        def __call__(self, *args: Any, **kwds: Any) -> Any:
            pass
    
    l = AsyncExceptionCollectorListenerWrapper(cast(AsyncExceptionCollectorListener, t_f12()))
    assert l.exception_type == BaseException
    assert l.exception_var_name is None
    assert l.accept_collector
    assert l.collector_var_name is None
    assert not l.reverse_order
    assert l.valid

    class t_f13:
        def __call__(self, es: AsyncExceptionCollector, /, a=True) -> Any:
            pass
    
    l = AsyncExceptionCollectorListenerWrapper(cast(AsyncExceptionCollectorListener, t_f13()))
    assert l.exception_type is None
    assert l.exception_var_name is None
    assert l.accept_collector
    assert l.collector_var_name is None
    assert l.valid
    
    ET = TypeVar('ET', bound=Exception)
    
    class t_f14(Generic[ET]):
        def __call__(self, a: ET) -> Any:
            pass
    l = AsyncExceptionCollectorListenerWrapper(cast(AsyncExceptionCollectorListener, t_f14[ValueError]()))
    assert l.exception_type == Exception
    assert l.exception_var_name == 'a'
    assert not l.accept_collector
    assert l.collector_var_name is None
    assert l.valid
    
    def t_f15():
        pass
    
    l = AsyncExceptionCollectorListenerWrapper(cast(AsyncExceptionCollectorListener, t_f15))
    assert l.exception_type is None
    assert l.exception_var_name is None
    assert not l.accept_collector
    assert l.collector_var_name is None
    assert not l.reverse_order
    assert l.valid
    
    def t_f16(a: int):
        pass

    try:
        l = AsyncExceptionCollectorListenerWrapper(cast(AsyncExceptionCollectorListener, t_f16))
    except ValueError:
        l = None
    assert not l
    
    def t_f17(a: SyntaxError, b: int):
        pass

    try:
        l = AsyncExceptionCollectorListenerWrapper(cast(AsyncExceptionCollectorListener, t_f17))
    except ValueError:
        l = None
    assert not l
    
    def t_f18(a: SyntaxError, b: int=0):
        pass

    l = AsyncExceptionCollectorListenerWrapper(cast(AsyncExceptionCollectorListener, t_f18))
    assert l.exception_type is SyntaxError
    assert l.exception_var_name == 'a'
    assert not l.accept_collector
    assert l.collector_var_name is None
    assert not l.reverse_order
    assert l.valid


import asyncio
import inspect
from dataclasses import dataclass
from typing import Any, Callable, Literal, TypeAlias, overload

from aiortc import RTCRtpTransceiver

from .codec import AudioEncodingParameters, VideoEncodingParameters
from .constants import AudioSourceInfo, VideoSourceInfo
from .mediaformat import Resolution, TrackKink
from .portal_types import VideoParameters as SigVideoParameters
from .sub_types import (AudioSubscriptionConstraints,
                        VideoSubscriptionConstraints)
from .webrtc import RTCRtpEncodingParameters
from .utils import ensure_future


class ConferenceError(Exception):
    message: str
    
    def __init__(self, message: str, *args: object) -> None:
        self.message = message
        super().__init__(self.message, *args)
        
    def mean_video_stream_not_exist(self):
        return self.message == "Target video stream does NOT satisfy:Requested video stream"
    
    def mean_audio_stream_not_exist(self):
        return self.message == "Target audio stream does NOT satisfy:Requested audio stream"

class OwtEvent:
    type: str
    
    def __init__(self, type: str) -> None:
        self.type = type
        
class ErrorEvent(OwtEvent):
    error: Exception
    
    def __init__(self, type: str, error: Exception) -> None:
        super().__init__(type)
        self.error = error
        
class MessageEvent(OwtEvent):
    
    def __init__(self, type: str, message: Any, origin: str | None = None, to: str | None = None) -> None:
        super().__init__(type)
        self.message = message
        self.origin = origin
        self.to = to
        
class MuteEvent(OwtEvent):
    kind: TrackKink
    
    def __init__(self, type: str, kind: TrackKink) -> None:
        super().__init__(type)
        self.kind = kind
    

Listener = Callable[[OwtEvent], Any]

@dataclass
class ListenerBox:
    listener: Listener
    once: bool
    seq: bool
    
class CompoundError(ConferenceError):
    errors: list[BaseException]
    
    def __init__(self, errors: list[BaseException]) -> None:
        super().__init__('|'.join((f'{e}' for e in errors)))
        self.errors = errors

class EventDispatcher:
    
    listeners_map: dict[str, list[ListenerBox]]
    lock: asyncio.Lock
    
    def __init__(self) -> None:
        self.listeners_map = {}
        self.lock = asyncio.Lock()
    
    @overload
    def addEventListener(self, eventType: Literal['data'], listener: Callable[[MessageEvent], Any], once: bool = False, seq: bool = True) -> None: ...
    @overload
    def addEventListener(self, eventType: Literal['id'], listener: Callable[[MessageEvent], Any], once: bool = False, seq: bool = True) -> None: ...
    @overload
    def addEventListener(self, eventType: Literal['ended'], listener: Callable[[OwtEvent], Any], once: bool = False, seq: bool = True) -> None: ...
    @overload
    def addEventListener(self, eventType: str, listener: Listener, once: bool = False, seq: bool = True) -> None: ...
    def addEventListener(self, eventType: str, listener: Any, once: bool = False, seq: bool = True) -> None:
        if eventType not in self.listeners_map:
            self.listeners_map[eventType] = []
        self.listeners_map[eventType].append(ListenerBox(listener=listener, once=once, seq=seq))
    
    def removeEventListener(self, eventType: str, listener: Any) -> None:
        if eventType in self.listeners_map:
            boxes = self.listeners_map[eventType]
            box = next((box for box in boxes if box.listener == listener), None)
            if box:
                boxes.remove(box)
                if len(boxes) == 0:
                    del self.listeners_map[eventType]
    
    def clearEventListener(self, eventType: str) -> None:
        if eventType in self.listeners_map:
            del self.listeners_map[eventType]
    
    async def dispatchEventAwaitable(self, event: OwtEvent):
        async with self.lock:
            boxes = self.listeners_map.get(event.type, [])
            if len(boxes) == 0:
                return
            pal_listeners: list[Listener] = []
            seq_listeners: list[Listener] = []
            for box in boxes:
                if box.seq:
                    seq_listeners.append(box.listener)
                else:
                    pal_listeners.append(box.listener)
            async def invoke_pal_listeners():
                results = await asyncio.gather(*[ensure_future(l, event) for l in pal_listeners], return_exceptions=True)
                excs = [e for e in results if isinstance(e, BaseException)]
                if len(excs) > 0:
                    if len(excs) > 1:
                        raise CompoundError(excs)
                    else:
                        raise excs[0]
            async def invoke_seq_listeners():
                for l in seq_listeners:
                    await ensure_future(l, event)
            results = await asyncio.gather(invoke_pal_listeners(), invoke_seq_listeners(), return_exceptions=True)
            boxes = [box for box in boxes if not box.once]
            self.listeners_map[event.type] = boxes
            errors: list[BaseException] = []
            for res in results:
                if isinstance(res, BaseException):
                    if isinstance(res, CompoundError):
                        errors.extend(res.errors)
                    else:
                        errors.append(res)
            if len(errors) > 0:
                if len(errors) > 1:
                    raise CompoundError(errors)
                else:
                    raise errors[0]
            
    def dispatchEvent(self, event: OwtEvent, loop: asyncio.AbstractEventLoop | None = None):
        asyncio.ensure_future(self.dispatchEventAwaitable(event), loop=loop)
        
                

@dataclass
class AudioTransceiver:
    type: Literal['audio']
    transceiver: RTCRtpTransceiver
    option: list[AudioEncodingParameters] | list[RTCRtpEncodingParameters] | AudioSubscriptionConstraints | VideoSubscriptionConstraints | bool
    source: AudioSourceInfo | None = None
    from_id: str | None = None
    parameters: None = None
    
@dataclass
class VideoParameters:
    resolution: Resolution | None = None
    framerate: float | None = None
    bitrate: str | float | None = None
    keyFrameInterval: float | None = None
    
    def to_dict(self) -> SigVideoParameters:
        return {
            'resolution': self.resolution.to_dict() if self.resolution else None,
            'framerate': self.framerate,
            'bitrate': self.bitrate,
            'keyFrameInterval': self.keyFrameInterval,
        }
    
@dataclass
class VideoTransceiver:
    type: Literal['video']
    transceiver: RTCRtpTransceiver
    option: list[VideoEncodingParameters] | list[RTCRtpEncodingParameters] | AudioSubscriptionConstraints | VideoSubscriptionConstraints | bool
    source: VideoSourceInfo | None = None
    from_id: str | None = None
    parameters: VideoParameters | None = None
    
Transceiver: TypeAlias = AudioTransceiver | VideoTransceiver


from dataclasses import dataclass
from typing import Any, Awaitable, Callable

from aiortc import MediaStreamTrack

from .common import EventDispatcher
from .mediaformat import Resolution, TrackKink
from .sub_types import (AudioSubscriptionConstraints,
                        VideoSubscriptionConstraints)
from .transport import TransportConstraints, TransportSettings
from .quic_types import QuicBidirectionalStream


@dataclass
class SubscribeOptions:
    audio: AudioSubscriptionConstraints | bool | None = None
    video: VideoSubscriptionConstraints | bool | None = None
    transport: TransportConstraints | None = None
    
@dataclass
class VideoSubscriptionUpdateOptions:
    resolution: Resolution | None = None
    frameRate: float | None = None
    bitrateMultiplier: float | None = None
    keyFrameInterval: float | None = None

@dataclass    
class SubscriptionUpdateOptions:
    video: VideoSubscriptionUpdateOptions | None = None

class Subscription(EventDispatcher):
    id: str
    tracks: list[MediaStreamTrack]
    quic_stream: QuicBidirectionalStream | None
    transport: TransportSettings
    stop: Callable[[], Awaitable[Any]]
    mute: Callable[[TrackKink], Awaitable[Any]]
    unmute: Callable[[TrackKink], Awaitable[Any]]
    _audioTrackId: str | None
    _videoTrackId: str | None
    
    
    def __init__(
        self, id: str,
        track: list[MediaStreamTrack],
        quic_stream: QuicBidirectionalStream | None,
        transport: TransportSettings,
        stop: Callable[[], Any],
        mute: Callable[[TrackKink], Any],
        unmute: Callable[[TrackKink], Any],
    ) -> None:
        super().__init__()
        self.id = id
        self.tracks = track
        self.quic_stream = quic_stream
        self.transport = transport
        self.stop = stop
        self.mute = mute
        self.unmute = unmute
        
    def audio_track(self) -> MediaStreamTrack | None:
        return next((t for t in self.tracks if t.kind == 'audio'), None)
    
    def video_track(self) -> MediaStreamTrack | None:
        return next((t for t in self.tracks if t.kind == 'video'), None)
    
    def tracks_ready(self) -> bool:
        return len(self.tracks) == len(self.transport.rtpTransceivers)
    
    def data_stream(self):
        if self.transport.type != 'quic':
            raise ConnectionError('Only data subscription has data stream atribute.')
        assert self.quic_stream
        return self.quic_stream
        
    

    


from dataclasses import dataclass, field

from .codec import AudioCodecParameters, VideoCodecParameters
from .mediaformat import Resolution

@dataclass
class AudioSubscriptionCapabilities:
    codecs: list[AudioCodecParameters] = field(default_factory=list)
    
@dataclass
class VideoSubscriptionCapabilities:
    codecs: list[VideoCodecParameters] = field(default_factory=list)
    resolutions: list[Resolution] = field(default_factory=list)
    frameRates: list[float] = field(default_factory=list)
    bitrateMultipliers: list[float] = field(default_factory=list)
    keyFrameIntervals: list[float] = field(default_factory=list)
    
@dataclass
class SubscriptionCapabilities:
    audio: AudioSubscriptionCapabilities | None = None
    video: VideoSubscriptionCapabilities | None = None
    
@dataclass
class AudioSubscriptionConstraints:
    codecs: list[AudioCodecParameters] | None = None
    
@dataclass
class VideoSubscriptionConstraints:
    codecs: list[VideoCodecParameters] | None = None
    resolution: Resolution | None = None
    frameRate: float | None = None
    bitrateMultiplier: float | None = None
    keyFrameInterval: float | None = None
    rid: str | None = None


import asyncio
from aioquic.asyncio import QuicConnectionProtocol
from aioquic.asyncio.client import connect, QuicConfiguration
from aioquic.h3.connection import H3Connection

import json
import base64
import logging
import struct
from urllib.parse import urlparse
from contextlib import AsyncExitStack
from typing import TypedDict

from .signaling import SioSignaling
from .common import EventDispatcher
from .stream import LocalStream, StreamSourceInfo, RemoteStream, OwtEvent
from .quic_types import QuicBidirectionalStream
from .common import ConferenceError
from .publication import Publication, TransportSettings
from .subscription import Subscription
from .log import configLogger

logger = configLogger(logger=logging.getLogger('owt-quic'))

class QuicToken(TypedDict):
    transportId: str

class QuicChannel(EventDispatcher):
    url: str
    tokenString: str
    signaling: SioSignaling
    configure: QuicConfiguration | None
    _transportId: str
    _loop: asyncio.AbstractEventLoop | None
    _ready: asyncio.Event
    _ready_lock: asyncio.Lock
    _lock: asyncio.Lock
    _quic: QuicConnectionProtocol | None
    _stack: AsyncExitStack | None
    _publications: dict[str, Publication]
    _subscriptions: dict[str, Subscription]
    _sub_promises: dict[str, asyncio.Future[Subscription]]
    _sub_streams: dict[str, QuicBidirectionalStream]
    
    def __init__(self, url: str, tokenString: str, signaling: SioSignaling, configure: QuicConfiguration | None = None, loop: asyncio.AbstractEventLoop | None = None) -> None:
        self.signaling = signaling
        self.url = url
        self.tokenString = tokenString
        self.configure = configure
        token: QuicToken = json.loads(base64.b64decode(tokenString))
        self._transportId = token['transportId']
        self._loop = loop
        self._lock = asyncio.Lock()
        self._ready = asyncio.Event()
        self._ready_lock = asyncio.Lock()
        self._publications = {}
        self._subscriptions = {}
        self._sub_streams = {}
        self._sub_promises = {}
        self._quic = None
        self._stack = None
        
    async def send_data(self, writer: asyncio.StreamWriter, data: str | bytes, drain: bool = True):
        if isinstance(data, str):
            content_bytes = data.encode(encoding='utf-8')
        else:
            content_bytes = data
        length = len(content_bytes)
        length_bytes = struct.pack('<I', length)
        writer.write(length_bytes)
        writer.write(content_bytes)
        if drain:
            await writer.drain()
    
    ZERO_BYTES: bytes = b'\0' * 16
    
    async def do_nothing(self, _):
        pass
    
    def _create_subscription(self, sub_id: str, stream: QuicBidirectionalStream) -> Subscription:
        async def stop():
            await self._unsubscribe(sub_id)
        return Subscription(
            id=sub_id,
            track=[],
            quic_stream=stream,
            transport=TransportSettings(type='quic', id=self._transportId, rtpTransceivers=[]),
            stop=stop,
            mute=self.do_nothing,
            unmute=self.do_nothing,
        )
            
    
    async def bind_subscription(self, stream: QuicBidirectionalStream):
        async with stream.lock:
            try:
                bytes = await stream.reader.readexactly(16)
            except asyncio.IncompleteReadError as e:
                logger.error(f'Stream closed unexpectedly.')
                return
        sub_id = bytes.hex()
        async with self._lock:
            assert sub_id not in self._subscriptions, f'Subscription with id {sub_id} exists.'
            assert sub_id not in self._sub_streams, f'Subscription stream with id {sub_id} exists.'
            self._sub_streams[sub_id] = stream
            future = self._sub_promises.get(sub_id)
        if future is not None and not future.done():
            del self._sub_promises[sub_id]
            subscription = self._create_subscription(sub_id, stream)
            self._subscriptions[sub_id] = subscription
            future.set_result(subscription)
        
    
    def on_stream_add(self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter):
        quic_stream = QuicBidirectionalStream(reader=reader, writer=writer, lock=asyncio.Lock())
        asyncio.ensure_future(self.bind_subscription(quic_stream), loop=self._loop)
    
    async def init(self):
        async with self._ready_lock:
            if self._ready.is_set():
                raise asyncio.InvalidStateError('quic connection exists!')
            parsed_result = urlparse(self.url)
            assert parsed_result.scheme in (
                "https",
                "wss",
            ), f"Only https:// or wss:// URLs are supported, but got {self.url}"
            host = parsed_result.hostname
            port = parsed_result.port or 7700
            assert host, f"no host: {self.url}"
            if self._quic:
                logger.warning('quic connection exists!')
                assert self._stack
                await self._stack.aclose()
                self._stack = None
                self._quic = None
            assert self._stack is None
            self._stack = AsyncExitStack()
            try:
                self._quic = await self._stack.enter_async_context(connect(host, port, configuration=self.configure, stream_handler=self.on_stream_add))
                reader, writer = await self._quic.create_stream()
                # 128 bit of zero indicates this is a stream for signaling.
                writer.write(self.ZERO_BYTES)
                await writer.drain()
                await self.send_data(writer, self.tokenString, False)
                # when signaling, server return 128 bit of zero again
                res = await reader.read(16)
                assert res == self.ZERO_BYTES
                self._ready.set()
                logger.debug('Authentication success.')
            except BaseException as e:
                if self._quic and self._quic._connected:
                    try:
                        self._quic.close()
                        await asyncio.wait_for(self._quic.wait_closed(), timeout=10)
                    except asyncio.TimeoutError:
                        logger.warning('Failed quic connection closed timeout.')
                        pass
                    except BaseException as e0:
                        logger.exception(e0)
                        pass
                raise e
        
    async def close(self, may_be_closed: bool = True):
        async with self._ready_lock:
            if not self._ready.is_set():
                if not may_be_closed:
                    raise asyncio.InvalidStateError('The quic channel has been closed.')
                else:
                    return
            publications = list(self._publications.values())
            for pub in publications:
                try:
                    await pub.stop()
                except BaseException as e:
                    logger.exception(e)
            publications.clear()
            subscriptions = list(self._subscriptions.values())
            for sub in subscriptions:
                try:
                    await sub.stop()
                except BaseException as e:
                    logger.exception(e)
            subscriptions.clear()
            for f in self._sub_promises.values():
                f.cancel()
            self._sub_promises.clear()
            self._sub_streams.clear()
            assert self._stack
            await self._stack.aclose()
            self._stack = None
            self._quic = None
            self._ready.clear()
            
    def is_ready(self):
        return self._ready.is_set()
    
    async def wait_ready(self):
        return await self._ready.wait()
            
    
    async def createSendStream(self, attributes: dict | None = None, is_unidirectional: bool = False) -> LocalStream:
        quic = self._quic
        if not quic:
            raise asyncio.InvalidStateError('The quic connection has not been inited or has been closed.')
        reader, writer = await quic.create_stream(is_unidirectional=is_unidirectional)
        quic_stream = QuicBidirectionalStream(reader=reader, writer=writer, lock=asyncio.Lock())
        return LocalStream(tracks=[], quic_stream=quic_stream, source=StreamSourceInfo(data=True), attributes=attributes or {})
    
    async def _unpublish(self, pub_id: str):
        self._publications.pop(pub_id, None)
        try:
            await self.signaling.unpublish(pub_id)
        except BaseException as e:
            logger.warning(f'MCU returns negative ack for unpublishing, {e}')
    
    async def publish(self, stream: LocalStream):
        if not stream.source.data:
            raise ConferenceError('Only data stream can be published by quic channel.')
        assert stream.quic_stream, 'The quic channel require the local stream has quic stream.'
        pub = await self.signaling.publish({
            'data': True,
            'transport': {
                'type': 'quic',
                'id': self._transportId,
            }
        })
        pub_id = pub['id']
        try:
            assert pub_id not in self._publications, f'The publication with id: {id} has exist.'
            if pub['transportId'] != self._transportId:
                raise ConferenceError(f'Transport ID not match.')
            writer = stream.quic_stream.writer
            id_bytes = bytes.fromhex(pub_id)
            assert len(id_bytes) == 16, f'Invalid publish id, it should be 128 bit hex string, but now is: {pub_id}'
            async with stream.quic_stream.lock:
                await self.send_data(writer, id_bytes)
            async def do_nothing(_):
                pass
            async def stop():
                self._publications.pop(pub_id, None)
                try:
                    await self.signaling.unpublish(pub_id)
                except BaseException as e:
                    logger.warning(f'MCU returns negative ack for unpublishing, {e}')
            publication = Publication(
                pub_id,
                transport=TransportSettings('quic', self._transportId, []),
                stop=stop,
                mute=do_nothing,
                unmute=do_nothing,
            )
            self._publications[pub_id] = publication
            return publication
        except BaseException:
            try:
                await self._unpublish(pub_id)
            except BaseException as e0:
                logger.warning(f'Unpublish failed. {e0}')
            raise
        
    async def _unsubscribe(self, sub_id: str):
        f = self._sub_promises.pop(sub_id, None)
        if f is not None:
            f.cancel()
        self._subscriptions.pop(sub_id, None)
        try:
            await self.signaling.unsubscribe(sub_id)
        except BaseException:
            logger.warning('MCU returns negative ack for unsubscribing, {e}')
        
    async def subscribe(self, stream: RemoteStream):
        if not stream.source.data:
            raise ConferenceError('Only data stream can be subscribed by quic channel.')
        assert stream.quic_stream, 'The quic channel require the remote stream has quic stream.'
        sub = await self.signaling.subscribe({
            'data': { 'from': stream.id },
            'transport': { 'type': 'quic', 'id': self._transportId },
        })
        
        sub_id = sub['id']
        try:
            assert sub_id not in self._subscriptions, f'The subscription with id: {id} has exist.'
            if sub['transportId'] != self._transportId:
                raise ConferenceError(f'Transport ID not match.')
            async with self._lock:
                quic_stream = self._sub_streams.get(sub_id, None)
                if quic_stream is not None:
                    subscription = self._create_subscription(sub_id, quic_stream)
                    self._subscriptions[sub_id] = subscription
                    return subscription
                else:
                    assert sub_id not in self._sub_promises
                    future: asyncio.Future[Subscription] = asyncio.Future(loop=self._loop)
                    self._sub_promises[sub_id] = future
            return await future
        except BaseException:
            try:
                await self._unsubscribe(sub_id)
            except BaseException as e0:
                logger.warning(f'unsubscribe failed. {e0}')
            raise


import hmac
from hashlib import sha256
import base64

class Auth:
    
    realm: str
    signature_method: str
    username: str
    role: str
    service_id: str
    service_key: str
    
    
    def __init__(self, service_id: str, service_key: str, username: str = '', role: str = '') -> None:
        self.realm = 'http://webrtc.intel.com'
        self.signature_method = 'HMAC_SHA256'
        self.service_id = service_id
        self.username = username
        self.role = role
        
    def hmac_sha256(self, to_sign: str, key: str):
        hash = hmac.new(base64.b64decode(key), to_sign.encode(), sha256)
        signed = base64.b64encode(hash.digest())
        return signed.decode()
    
    def sign(self, timestamp: int, nonce: int) -> str:
        to_sign = f'{timestamp},{nonce}'
        if self.username and self.role:
            to_sign += f',{self.username},{self.role}'
        return self.hmac_sha256(to_sign, self.service_key)


from collections.abc import Mapping, MutableMapping
from typing import Any
import os
import logging

def getLevel(name: str | None) -> str | int | None:
    level = None
    if name:
        level = os.environ.get(f'LOGGER_{name.upper().replace("-", "_")}_LEVEL')
    if not level:
        level = os.environ.get(f'LOGGER_OWT_LEVEL', default=os.environ.get(f'LOGGER_LEVEL'))
    if level:
        if level.isdigit():
            level = int(level)
        else:
            level = level.upper()
    return level
        

def configRoot():
    format = os.environ.get('LOGGER_FORMAT', '%(asctime)s - %(filename)s:%(funcName)s [%(levelname)s] [%(process)d-%(thread)d] [%(name)s] %(message)s')
    level = getLevel(None)
    logging.basicConfig(format=format, level=level)
    fh = logging.FileHandler('ai.log')
    logging.root.addHandler(fh)

def configLogger(logger: logging.Logger) -> logging.Logger:
    level = getLevel(logger.name)
    if level is not None:
        logger.setLevel(level)
    return logger
    

class MyLoggerAdapter(logging.LoggerAdapter):
    
    prefix: str
    
    def __init__(self, logger: logging.Logger, prefix: str | None = None, extra: Mapping[str, object] | None = None) -> None:
        super().__init__(logger, extra)
        self.prefix = prefix or ''
        
    def process(self, msg: Any, kwargs: MutableMapping[str, Any]) -> tuple[Any, MutableMapping[str, Any]]:
        if 'extra' not in kwargs:
            kwargs["extra"] = self.extra

        return f'{self.prefix}{msg}', kwargs


import asyncio
import base64
import json
import logging
import time
from typing import Any, Literal, overload

import socketio

from .common import ConferenceError, EventDispatcher, MessageEvent, OwtEvent
from .portal_types import (BWEResult, CandidateMessage, LoginInfo, LoginResult,
                           OfferAnswer, PublicationOrSubscriptionResult,
                           PublicationRequest, ReconnectionTicket,
                           ReloginResponse, RemovedCandidatesMessage,
                           SOACMessage, StreamControlInfo,
                           SubscriptionControlInfo, SubscriptionRequest,
                           TextSendMessage, UnpublicationRequest,
                           UnsubscriptionRequest)
from .utils import AsyncExceptionCollector, NullAsyncContext
from .log import configLogger

logger = configLogger(logger=logging.getLogger("owt-signaling"))

class SioSignaling(EventDispatcher):
    
    io: socketio.AsyncClient | None
    _lock: asyncio.Lock
    _loop: asyncio.AbstractEventLoop | None
    _ready_future: asyncio.Future[None]
    _loggedIn: bool = False
    _reconnectTimes: int = 0
    _reconnectionTicket: str | None = None
    _refreshReconnectionTicket: asyncio.Future | None = None
    reconnectionAttempts: int
    exception_collector: AsyncExceptionCollector | None = None
    
    def __init__(self, loop: asyncio.AbstractEventLoop | None = None) -> None:
        super().__init__()
        self.io = None
        self._loop = loop
        self._lock = asyncio.Lock()
        self._ready_future = asyncio.Future(loop=loop)
        
    async def ready(self):
        async with self._lock:
            await self._ready_future
            
    async def markReady(self):
        if self.io is None:
            return
        async with self._lock:
            if not self._ready_future.done():
                self._ready_future.set_result(None)
                self._loggedIn = True
            
    async def markNotReady(self):
        if self.io is None:
            return
        async with self._lock:
            if self.io and self._ready_future.done():
                self._ready_future = asyncio.Future()
                self._loggedIn = False
                
    async def cancelReady(self, e: BaseException | None = None):
        async with self._lock:
            if not self._ready_future.done():
                if e and not isinstance(e, asyncio.CancelledError):
                    self._ready_future.set_exception(e)
                else:
                    self._ready_future.cancel()
            else:
                self._ready_future.cancel()
                    
    
        
    async def connect(self, host: str, login_info: LoginInfo) -> LoginResult:
        if self.io is not None:
            raise RuntimeError('Portal has been connected.')
        io = self.io = socketio.AsyncClient(ssl_verify=False)
        collector = self.exception_collector = AsyncExceptionCollector(loop=self._loop)
        if not login_info.get('reconnection'):
            login_info['reconnection'] = {
                'keepTime': 300
            }
        async def on_exception(e: BaseException):
            await self.disconnect(True, True, e)
            
        await collector.add_listener(on_exception)
        
        await io.connect(host)
        await self.dispatchEventAwaitable(OwtEvent('connected'))
        def create_message_handler(notification: str):
            async def message_handler(data: Any):
                async with collector.collect():
                    await self.dispatchEventAwaitable(MessageEvent(
                        'data',
                        message={
                            'notification': notification,
                            'data': data,
                        }
                    ))
            return message_handler
    
        for notification in ['participant', 'text', 'stream', 'progress']:
            io.on(notification, create_message_handler(notification))
            
        async def on_drop(data: Any):
            async with collector.collect():
                await self.disconnect(False, True)
        io.on('drop', on_drop)
        
        async def on_disconnect():
            async with collector.collect():
                logger.debug('signaling disconnected')
                await self.markNotReady()
                # self._clearReconnectionTask()
        io.on('disconnect', on_disconnect)
            
        login_result = await self.send('login', login_info)
        await self.markReady()
        reconnectionTicket = login_result.get('reconnectionTicket')
        assert reconnectionTicket is not None
        self._onReconnectionTicket(reconnectionTicket)
        
        async def on_connect():
            async with collector.collect():
                logger.debug('signaling reconnected')
                await self.dispatchEventAwaitable(OwtEvent('reconnected'))
                try:
                    assert self._reconnectionTicket is not None
                    relogin_resp = await self.send('relogin', self._reconnectionTicket)
                    await self.markReady()
                    self._onReconnectionTicket(relogin_resp['ticket'])
                except BaseException as e:
                    logger.error(f'Something error happened when relogining. {e}')
                    raise e

        io.on('connect', on_connect)
        return login_result

    @overload
    async def send(self, requestName: Literal['soac'], requestData: SOACMessage) -> None: ...
    @overload
    async def send(self, requestName: Literal['publish'], requestData: PublicationRequest) -> PublicationOrSubscriptionResult: ...
    @overload
    async def send(self, requestName: Literal['unpublish'], requestData: UnpublicationRequest) -> None: ...
    @overload
    async def send(self, requestName: Literal['stream-control'], requestData: StreamControlInfo) -> None: ...
    @overload
    async def send(self, requestName: Literal['subscribe'], requestData: SubscriptionRequest) -> PublicationOrSubscriptionResult: ...
    @overload
    async def send(self, requestName: Literal['unsubscribe'], requestData: UnsubscriptionRequest) -> None: ...
    @overload
    async def send(self, requestName: Literal['subscription-control'], requestData: SubscriptionControlInfo) -> BWEResult | None: ...
    @overload
    async def send(self, requestName: Literal['login'], requestData: LoginInfo) -> LoginResult: ...
    @overload
    async def send(self, requestName: Literal['relogin'], requestData: str) -> ReloginResponse: ...
    @overload
    async def send(self, requestName: Literal['refreshReconnectionTicket']) -> str: ...
    @overload
    async def send(self, requestName: Literal['text'], requestData: TextSendMessage) -> None: ...
    async def send(self, requestName: str, requestData: Any = None):
        if self.io is None:
            raise ConferenceError('Portal is not connected.')
        future = asyncio.Future(loop=self._loop)
        def resp(status: str, data: Any = None):
            if status == 'ok' or status == 'success':
                future.set_result(data)
            else:
                future.set_exception(ConferenceError(data))
        await self.io.emit(requestName, requestData, callback=resp)
        return await future
    
    async def send_text(self, message: str, to: str = 'all'):
        await self.ready()
        return await self.send('text', { 'message': message, 'to': to })
    
    async def publish(self, request: PublicationRequest) -> PublicationOrSubscriptionResult:
        await self.ready()
        return await self.send('publish', request)
    
    async def unpublish(self, pub_id: str, maybe_not_exist: bool = True) -> None:
        await self.ready()
        try:
            return await self.send('unpublish', { 'id': pub_id })
        except ConferenceError as e:
            if e.message and e.message.lower() != 'Stream does NOT exist'.lower() or not maybe_not_exist:
                raise e
        
    async def subscribe(self, request: SubscriptionRequest) -> PublicationOrSubscriptionResult:
        await self.ready()
        return await self.send('subscribe', request)
    
    async def unsubscribe(self, sub_id: str, maybe_not_exist: bool = True):
        await self.ready()
        try:
            await self.send('unsubscribe', { 'id': sub_id })
        except ConferenceError as e:
            if e.message and e.message.lower() != 'Subscription does NOT exist'.lower() or not maybe_not_exist:
                raise e
            
    async def soac(self, transport_id: str, signaling: OfferAnswer | CandidateMessage | RemovedCandidatesMessage) -> None:
        await self.ready()
        await self.send('soac', { 'id': transport_id, 'signaling': signaling })    
    
    async def disconnect(self, logout: bool = True, permit_duplicate_disconnect = False, e: BaseException | None = None):
        if self.io is None:
            if not permit_duplicate_disconnect:
                raise asyncio.InvalidStateError('Portal is not connected.')
            else:
                return
        async with self.exception_collector.collect() if self.exception_collector else NullAsyncContext():
            await self.cancelReady(e)
            io = self.io
            self.io = None
            if io.connected:
                if logout:
                    try:
                        await io.emit('logout')
                    except BaseException as e:
                        logger.warning(f'Portal logout failed: {e}')
                self._clearReconnectionTask()
                try:
                    await io.disconnect()
                except BaseException as e:
                    logger.warning(f'IO socket disconnect failed: {e}')
            await self.dispatchEventAwaitable(OwtEvent('disconnect'))
            if self.exception_collector:
                self.exception_collector.complete()
                self.exception_collector = None

        
    def _onReconnectionTicket(self, ticketString: str):
        self._reconnectionTicket = ticketString
        ticket: ReconnectionTicket = json.loads(base64.b64decode(ticketString))
        
        # Refresh ticket 1 min or 10 seconds before it expires.
        now = time.time() * 1000
        millisecondsInOneMinute = 60 * 1000
        millisecondsInTenSeconds = 10 * 1000
        if ticket['notAfter'] <= now - millisecondsInTenSeconds:
            logger.warning('Reconnection ticket expires too soon.')
            return
        
        refreshAfter = ticket['notAfter'] - now - millisecondsInOneMinute
        if refreshAfter < 0:
            refreshAfter = ticket['notAfter'] - now - millisecondsInTenSeconds
        
        self._clearReconnectionTask()
        
        async def refreshReconnectionTicket():
            await asyncio.sleep(refreshAfter / 1000)
            io = self.io
            if io is not None:
                new_reconnect_ticket = await self.send('refreshReconnectionTicket')
                self._onReconnectionTicket(new_reconnect_ticket)
        self._refreshReconnectionTicket = asyncio.ensure_future(refreshReconnectionTicket(), loop=self._loop)
    
    def _clearReconnectionTask(self):
        if self._refreshReconnectionTicket is not None:
            self._refreshReconnectionTicket.cancel()
        self._refreshReconnectionTicket = None



from typing import Any, Callable, Literal, overload

from .common import EventDispatcher, OwtEvent


class Participant(EventDispatcher):
    id: str
    """
    The ID of the participant. It varies when a single user join different conferences.
    """
    role: str
    userId: str
    """
    The user ID of the participant. It can be integrated into existing account management system.
    """
    def __init__(self, id: str, role: str, userId: str) -> None:
        super().__init__()
        self.id = id
        self.role = role
        self.userId = userId

    @overload
    def addEventListener(self, eventType: Literal['left'], listener: Callable[[OwtEvent], Any], once: bool = False, seq: bool = True) -> None: ...
    @overload
    def addEventListener(self, eventType: str, listener: Callable[[OwtEvent], Any], once: bool = False, seq: bool = True) -> None: ...
    def addEventListener(self, eventType: str, listener: Callable[[OwtEvent], Any], once: bool = False, seq: bool = True) -> None:
        return super().addEventListener(eventType, listener, once=once, seq=seq)